@@ -9,15 +9,29 @@
 #
 
 
-import os, sys, re
+import datetime
+import os 
+import sys
+import regex as re
 import inspect
+import pkg_resources
 import subprocess
 import tempfile
+from pathlib import Path
 
 # These are features added to SolidPython but NOT in OpenSCAD.
 # Mark them for special treatment
 non_rendered_classes = ['hole', 'part']
 
+# Words reserved in Python but not OpenSCAD
+# Re: https://github.com/SolidCode/SolidPython/issues/99
+PYTHON_ONLY_RESERVED_WORDS = [
+    'False','class','finally','is','return','None','continue','lambda',
+    'try','True','def','from','nonlocal','while','and','del','global','not',
+    'with','as','elif','or','yield','assert','pass', 'break','except','in','raise',
+]
+
+
 # =========================================
 # = Rendering Python code to OpenSCAD code=
 # =========================================
@@ -74,7 +88,7 @@ def scad_render_animated(func_to_animate, steps=20, back_and_forth=True, filepat
     # should avoid any rounding error problems, and doesn't require the file
     # to be animated with an identical number of steps to the way it was
     # created. -ETJ 28 Mar 2013
-    scad_obj = func_to_animate()
+    scad_obj = func_to_animate(0)
     include_strings = _find_include_strings(scad_obj)
     # and render the string
     includes = ''.join(include_strings) + "\n"
@@ -95,7 +109,7 @@ def scad_render_animated(func_to_animate, steps=20, back_and_forth=True, filepat
                 eval_time = time * 2
             else:
                 eval_time = 2 - 2 * time
-        scad_obj = func_to_animate(_time=eval_time)
+        scad_obj = func_to_animate(eval_time)
 
         scad_str = indent(scad_obj._render())
         rendered_string += ("if ($t >= %(time)s && $t < %(end_time)s){"
@@ -112,13 +126,19 @@ def scad_render_animated_file(func_to_animate, steps=20, back_and_forth=True,
 
 
 def scad_render_to_file(scad_object, filepath=None, file_header='', include_orig_code=True):
-    rendered_string = scad_render(scad_object, file_header)
+
+    header = "// Generated by SolidPython {version} on {date}\n".format(
+                 version = _get_version(),
+                 date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
+    header += file_header
+
+    rendered_string = scad_render(scad_object, header)
     return _write_code_to_file(rendered_string, filepath, include_orig_code)
 
 
 def _write_code_to_file(rendered_string, filepath=None, include_orig_code=True):
     try:
-        calling_file = os.path.abspath(calling_module(stack_depth=3).__file__)
+        calling_file = Path(calling_module(stack_depth=3).__file__).absolute()
 
         if include_orig_code:
             rendered_string += sp_code_in_scad_comment(calling_file)
@@ -128,27 +148,46 @@ def _write_code_to_file(rendered_string, filepath=None, include_orig_code=True):
         # If filepath isn't supplied, place a .scad file with the same name
         # as the calling module next to it
         if not filepath:
-            filepath = os.path.splitext(calling_file)[0] + '.scad'
+            filepath = filepath.with_ext('.scad')
     except AttributeError as e:
         # If no calling_file was found, this is being called from the terminal.
         # We can't read original code from a file, so don't try,
         # and can't read filename from the calling file either, so just save to
         # solid.scad.
         if not filepath:
-            filepath = os.path.abspath('.') + "/solid.scad"
+            filepath = Path().absolute() / 'solid.scad'
 
-    f = open(filepath, "w")
-    f.write(rendered_string)
-    f.close()
+    Path(filepath).write_text(rendered_string)
     return True
 
 
+def _get_version():
+    # Returns SolidPython version
+    # Raises a RuntimeError if the version cannot be determined
+
+    try:
+        # if SolidPython is installed use `pkg_resources`
+        return pkg_resources.get_distribution('solidpython').version
+
+    except pkg_resources.DistributionNotFound:
+        # if the running SolidPython is not the one installed via pip,
+        # try to read it from the project setup file
+        version_pattern = re.compile(r"version = ['\"]([^'\"]*)['\"]")
+        version_file_path = Path(__file__).parent.parent / 'pyproject.toml'
+
+        version_match = version_pattern.search(version_file_path.read_text())
+        if version_match:
+            return version_match.group(1)
+
+    raise RuntimeError("Unable to determine software version.")
+
+
 def sp_code_in_scad_comment(calling_file):
     # Once a SCAD file has been created, it's difficult to reconstruct
     # how it got there, since it has no variables, modules, etc.  So, include
     # the Python code that generated the scad code as comments at the end of
     # the SCAD code
-    pyopenscad_str = open(calling_file, 'r').read()
+    pyopenscad_str = Path(calling_file).read_text()
 
     # TODO: optimally, this would also include a version number and
     # git hash (& date & github URL?) for the version of solidpython used
@@ -169,8 +208,7 @@ def sp_code_in_scad_comment(calling_file):
 # = Parsing =
 # ===========
 def extract_callable_signatures(scad_file_path):
-    with open(scad_file_path) as f:
-        scad_code_str = f.read()
+    scad_code_str = Path(scad_file_path).read_text()
     return parse_scad_callables(scad_code_str)
 
 def parse_scad_callables(scad_code_str):
@@ -185,17 +223,17 @@ def parse_scad_callables(scad_code_str):
 
     # TODO:  write a pyparsing grammar for OpenSCAD, or, even better, use the yacc parse grammar
     # used by the language itself.  -ETJ 06 Feb 2011
-
+    
+    # FIXME: OpenSCAD use/import includes top level variables. We should parse 
+    # those out (e.g. x = someValue;) as well -ETJ 21 May 2019
     no_comments_re = r'(?mxs)(//.*?\n|/\*.*?\*/)'
 
     # Also note: this accepts: 'module x(arg) =' and 'function y(arg) {', both
     # of which are incorrect syntax
     mod_re = r'(?mxs)^\s*(?:module|function)\s+(?P<callable_name>\w+)\s*\((?P<all_args>.*?)\)\s*(?:{|=)'
 
-    # This is brittle.  To get a generally applicable expression for all arguments,
-    # we'd need a real parser to handle nested-list default args or parenthesized statements.
-    # For the moment, assume a maximum of one square-bracket-delimited list
-    args_re = r'(?mxs)(?P<arg_name>\w+)(?:\s*=\s*(?P<default_val>[\w.-]+|\[.*\]))?(?:,|$)'
+    # See https://github.com/SolidCode/SolidPython/issues/95; Thanks to https://github.com/Torlos
+    args_re = r'(?mxs)(?P<arg_name>\w+)(?:\s*=\s*(?P<default_val>([\w.\"\s\?:\-+\\\/*]+|\((?>[^()]|(?2))*\)|\[(?>[^\[\]]|(?2))*\])+))?(?:,|$)'
 
     # remove all comments from SCAD code
     scad_code_str = re.sub(no_comments_re, '', scad_code_str)
@@ -246,6 +284,12 @@ def new_openscad_class_str(class_name, args=[], kwargs=[], include_file_path=Non
     args_str = ''
     args_pairs = ''
 
+    # Re: https://github.com/SolidCode/SolidPython/issues/99
+    # Don't allow any reserved words as argument names or module names
+    # (They might be valid OpenSCAD argument names, but not in Python)
+    class_name = _subbed_keyword(class_name)
+
+    args = map(_subbed_keyword, args)
     for arg in args:
         args_str += ', ' + arg
         args_pairs += "'%(arg)s':%(arg)s, " % vars()
@@ -253,24 +297,23 @@ def new_openscad_class_str(class_name, args=[], kwargs=[], include_file_path=Non
     # kwargs have a default value defined in their SCAD versions.  We don't
     # care what that default value will be (SCAD will take care of that), just
     # that one is defined.
+    kwargs = map(_subbed_keyword, kwargs)
     for kwarg in kwargs:
         args_str += ', %(kwarg)s=None' % vars()
         args_pairs += "'%(kwarg)s':%(kwarg)s, " % vars()
 
     if include_file_path:
-        # include_file_path may include backslashes on Windows; escape them
-        # again here so any backslashes don't get used as escape characters
-        # themselves
-        include_file_path = include_file_path.replace('\\', '\\\\')
+        include_file_str = Path(include_file_path).as_posix()
 
         # NOTE the explicit import of 'solid' below. This is a fix for:
         # https://github.com/SolidCode/SolidPython/issues/20 -ETJ 16 Jan 2014
         result = ("import solid\n"
                   "class %(class_name)s(solid.IncludedOpenSCADObject):\n"
                   "   def __init__(self%(args_str)s, **kwargs):\n"
-                  "       solid.IncludedOpenSCADObject.__init__(self, '%(class_name)s', {%(args_pairs)s }, include_file_path='%(include_file_path)s', use_not_include=%(use_not_include)s, **kwargs )\n"
+                  "       solid.IncludedOpenSCADObject.__init__(self, '%(class_name)s', {%(args_pairs)s }, include_file_path='%(include_file_str)s', use_not_include=%(use_not_include)s, **kwargs )\n"
                   "   \n"
                   "\n" % vars())
+
     else:
         result = ("class %(class_name)s(OpenSCADObject):\n"
                   "   def __init__(self%(args_str)s):\n"
@@ -280,6 +323,23 @@ def new_openscad_class_str(class_name, args=[], kwargs=[], include_file_path=Non
 
     return result
 
+def _subbed_keyword(keyword):
+    # Append an underscore to any python reserved word. 
+    # No-op for all other strings, e.g. 'or' => 'or_', 'other' => 'other'
+    new_key = keyword + '_' if keyword in PYTHON_ONLY_RESERVED_WORDS else keyword
+    if new_key != keyword:
+        print("\nFound OpenSCAD code that's not compatible with Python. \n"
+              "Imported OpenSCAD code using `%s` \n"
+              "can be accessed with `%s` in SolidPython\n"%(keyword, new_key))
+    return new_key
+
+def _unsubbed_keyword(subbed_keyword):
+    # Remove trailing underscore for already-subbed python reserved words. 
+    # No-op for all other strings: e.g. 'or_' => 'or', 'other_' => 'other_'
+    shortened = subbed_keyword[:-1]
+    return shortened if shortened in PYTHON_ONLY_RESERVED_WORDS else subbed_keyword
+
+
 # =========================
 # = Internal Utilities    =
 # =========================
@@ -385,9 +445,16 @@ class OpenSCADObject(object):
         return s
 
     def _render_str_no_children(self):
-        s = "\n" + self.modifier + self.name + "("
+        callable_name = _unsubbed_keyword(self.name)
+        s = "\n" + self.modifier + callable_name + "("
         first = True
 
+        # Re: https://github.com/SolidCode/SolidPython/issues/99
+        # OpenSCAD will accept Python reserved words as callables or argument names,
+        # but they won't compile in Python. Those have already been substituted
+        # out (e.g 'or' => 'or_'). Sub them back here.
+        self.params = {_unsubbed_keyword(k):v for k, v in self.params.items()}
+
         # OpenSCAD doesn't have a 'segments' argument, but it does
         # have '$fn'.  Swap one for the other
         if 'segments' in self.params:
@@ -407,7 +474,7 @@ class OpenSCADObject(object):
 
         for k in all_params_sorted:
             v = self.params[k]
-            if v == None:
+            if v is None:
                 continue
 
             if not first:
@@ -432,31 +499,33 @@ class OpenSCADObject(object):
             if child.is_hole:
                 s += child._render(render_holes=True)
             elif child.has_hole_children:
-                # Holes exist in the compiled tree in two pieces:
-                # The shapes of the holes themselves, (an object for which
-                # obj.is_hole is True, and all its children) and the
-                # transforms necessary to put that hole in place, which
-                # are inherited from non-hole geometry.
-
-                # Non-hole Intersections & differences can change (shrink)
-                # the size of holes, and that shouldn't happen: an
-                # intersection/difference with an empty space should be the
-                # entirety of the empty space.
-                #  In fact, the intersection of two empty spaces should be
-                # everything contained in both of them:  their union.
-                # So... replace all super-hole intersection/diff transforms
-                # with union in the hole segment of the compiled tree.
-                # And if you figure out a better way to explain this,
-                # please, please do... because I think this works, but I
-                # also think my rationale is shaky and imprecise. 
-                # -ETJ 19 Feb 2013
-                s = s.replace("intersection", "union")
-                s = s.replace("difference", "union")
                 s += child._render_hole_children()
         if self.name in non_rendered_classes:
             pass
         else:
             s = self._render_str_no_children() + "{" + indent(s) + "\n}"
+            
+        # Holes exist in the compiled tree in two pieces:
+        # The shapes of the holes themselves, (an object for which
+        # obj.is_hole is True, and all its children) and the
+        # transforms necessary to put that hole in place, which
+        # are inherited from non-hole geometry.
+
+        # Non-hole Intersections & differences can change (shrink)
+        # the size of holes, and that shouldn't happen: an
+        # intersection/difference with an empty space should be the
+        # entirety of the empty space.
+        #  In fact, the intersection of two empty spaces should be
+        # everything contained in both of them:  their union.
+        # So... replace all super-hole intersection/diff transforms
+        # with union in the hole segment of the compiled tree.
+        # And if you figure out a better way to explain this,
+        # please, please do... because I think this works, but I
+        # also think my rationale is shaky and imprecise. 
+        # -ETJ 19 Feb 2013
+        s = s.replace("intersection", "union")
+        s = s.replace("difference", "union")
+            
         return s
 
     def add(self, child):
@@ -473,6 +542,10 @@ class OpenSCADObject(object):
             if len(child) == 1 and isinstance(child[0], (list, tuple)):
                 child = child[0]
             [self.add(c) for c in child]
+        elif isinstance(child, int):
+            # Allowing for creating object by adding to 0 (as in sum())
+            if child != 0:
+                raise ValueError
         else:
             self.children.append(child)
             child.set_parent(self)
@@ -529,6 +602,13 @@ class OpenSCADObject(object):
         '''
         return objects.union()(self, x)
 
+    def __radd__(self, x):
+        '''
+        This makes u = a+b identical to:
+        u = union()(a, b )
+        '''
+        return objects.union()(self, x)
+
     def __sub__(self, x):
         '''
         This makes u = a - b identical to:
@@ -580,10 +660,8 @@ class IncludedOpenSCADObject(OpenSCADObject):
     def __init__(self, name, params, include_file_path, use_not_include=False, **kwargs):
         self.include_file_path = self._get_include_path(include_file_path)
 
-        if use_not_include:
-            self.include_string = 'use <%s>\n' % self.include_file_path
-        else:
-            self.include_string = 'include <%s>\n' % self.include_file_path
+        use_str = 'use' if use_not_include else 'include'
+        self.include_string = f'{use_str} <{self.include_file_path}>\n'
 
         # Just pass any extra arguments straight on to OpenSCAD; it'll accept
         # them
@@ -595,17 +673,18 @@ class IncludedOpenSCADObject(OpenSCADObject):
     def _get_include_path(self, include_file_path):
         # Look through sys.path for anyplace we can find a valid file ending
         # in include_file_path.  Return that absolute path
-        if os.path.isabs(include_file_path) and os.path.isfile(include_file_path):
-            return include_file_path
+        ifp = Path(include_file_path)
+        if ifp.is_absolute() and ifp.is_file():
+            return ifp
         else:
             for p in sys.path:
-                whole_path = os.path.join(p, include_file_path)
-                if os.path.isfile(whole_path):
-                    return os.path.abspath(whole_path)
+                path = Path(p)
+                whole_path = path / ifp
+                if whole_path.is_file():
+                    return whole_path.absolute()
 
         # No loadable SCAD file was found in sys.path.  Raise an error
-        raise ValueError("Unable to find included SCAD file: "
-                         "%(include_file_path)s in sys.path" % vars())
+        raise ValueError(f"Unable to find included SCAD file: {include_file_path} in sys.path")
 
 # now that we have the base class defined, we can do a circular import
 from . import objects
@@ -615,7 +694,12 @@ def py2openscad(o):
         return str(o).lower()
     if type(o) == float:
         return "%.10f" % o
-    if type(o) == list or type(o) == tuple:
+    if type(o) == str:
+        return '"' + o + '"'
+    if type(o).__name__ == "ndarray":
+        import numpy
+        return numpy.array2string(o, separator=",", threshold=1000000000)
+    if hasattr(o, "__iter__"):
         s = "["
         first = True
         for i in o:
@@ -625,8 +709,6 @@ def py2openscad(o):
             s += py2openscad(i)
         s += "]"
         return s
-    if type(o) == str:
-        return '"' + o + '"'
     return str(o)
 
 
